/*
  CRF++ -- Yet Another CRF toolkit
 
  $Id: crf_learn.cpp,v 1.9 2005/10/30 03:54:15 taku Exp $;

  Copyright (C) 2005 Taku Kudo <taku@chasen.org>

  This is free software with ABSOLUTELY NO WARRANTY.
  
  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.
  
  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of

  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.
  
  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#include "encoder.h"
#include "param.h"
#include <vector>
#include <string>
#include <cstring>
#include <iostream>

# include "ConfigFile.h"


int main (int argc, char **argv)
{

  CRFPP::Param param;
  static const CRFPP::Option long_options[] = 
    {
      {"config", 'c', "crf.conf", "STRING", "file to read configuration" },
      {"version",  'v', 0,        0,       "show the version and exit" },
      {"help",     'h', 0,        0,       "show this help and exit" },
      {0,0,0,0,0}
    };
  
  param.open (argc, argv, long_options);

  if (param.getProfileInt("version")) {
    std::cout << param.version (long_options);
    return 0;
  }
  
  if (param.getProfileInt("help")) {
    std::cout << param.help (long_options);
    return 0;
  }

  ConfigFile config(param.getProfileString("config"));
  
  CRFPP::Encoder encoder;
  bool convert;
  config.readInto(convert, "convert");
  
  if (convert) {
    string txtfile;
    config.readInto(txtfile, "txtfile");
    string binfile;
    config.readInto(binfile, "binfile");
    
    if (! encoder.convert(txtfile.c_str(), binfile.c_str())) {
      std::cerr << encoder.what() << std::endl;
      return -1;
    }
  } else {  
    if (! encoder.learn (config)) {
      std::cerr << encoder.what() << std::endl;
      return -1;
    }
  }
  return 0;
}
