/*
  CRF++ -- Yet Another CRF toolkit

  $Id: crf_test.cpp,v 1.7 2005/10/30 03:54:15 taku Exp $;

  Copyright (C) 2005 Taku Kudo <taku@chasen.org>

  This is free software with ABSOLUTELY NO WARRANTY.

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307	USA
*/
#include <string>
#include <vector>
#include "param.h"
#include "tagger.h"
#include "stream_wrapper.h"

int main (int argc, char **argv)
{
  static const CRFPP::Option long_options[] =
    {
      {"model",	 'm',  0,	"FILE",	 "set FILE for model file"},
      {"nbest",	 'n', "0",	"INT",	 "output n-best results"},
      {"verbose" , 'v', "0",	"INT",	 "set INT for verbose level"},
      {"output",	 'o',  0,	"FILE",	 "use FILE as output file"},
      {"version",	 'v',  0,	 0,	  "show the version and exit" },
      {"help",	 'h',  0,	 0,	  "show this help and exit" },
      {0,0,0,0,0}
    };

  CRFPP::Param param;

  param.open (argc, argv, long_options);

  std::string model = param.getProfileString ("model");
  int vlevel	    = param.getProfileInt ("verbose");
  size_t nbest	    = param.getProfileInt ("nbest");

  if (param.getProfileInt("version")) {
    std::cout <<  param.version (long_options);
    return -1;
  }

  if (param.getProfileInt("help") || model.empty()) {
    std::cout <<  param.help (long_options);
    return -1;
  }

  CRFPP::Tagger tagger;   
  if (! tagger.open (model.c_str())) {
    std::cerr << tagger.what() << std::endl;
    return -1;
  }

  // output 
  std::string output = param.getProfileString ("output");
  if (output.empty()) output = "-";
  CRFPP::ostream_wrapper os (output.c_str());
  if (! *os) {
    std::cerr << "open failed: " << output << std::endl;
    return -1;
  }

  // input
  const std::vector <std::string>& rest_ = param.getRestArg ();
  std::vector<std::string> rest = rest_; // trivial copy
  if (rest.size() == 0) rest.push_back("-"); 

  for (size_t i = 0; i < rest.size(); i++) {
    CRFPP::istream_wrapper is (rest[i].c_str());
    if (! *is) {
      std::cerr << "open failed: " << rest[i] << std::endl;
      return -1;
    }
    while (*is) tagger.parse (nbest, vlevel, *is, *os);
  }
   
  return 0;
}
