/*
  CRF++ -- Yet Another CRF toolkit

  $Id: encoder.cpp,v 1.10 2005/10/30 03:54:15 taku Exp $;

  Copyright (C) 2005 Taku Kudo <taku@chasen.org>

  This is free software with ABSOLUTELY NO WARRANTY.

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#include "encoder.h"
#include "timer.h"
#include "tagger.h"
#include "lbfgs.h"
#include "common.h"
#include "stream_wrapper.h"
#include <fstream>
#include <stdlib.h>
#include <iterator>
#include <cstring>
#include <cassert>
#include <algorithm>

#define WHAT_ERROR(msg) {                                               \
    WHAT(msg);                                                          \
    for (std::vector<Tagger *>::iterator it = x.begin(); it != x.end(); ++it) \
      delete *it;                                                       \
    x.clear();                                                          \
    return false; }


namespace CRFPP {

  bool Encoder::convert(const char* textfilename, const char *binaryfilename)
  {
    EncoderFeatureIndex feature_index;
    if (! feature_index.convert(textfilename, binaryfilename))
      WHAT_FALSE(feature_index.what());

    return true;
  }

  std::string Encoder::itostr(int i) {
    
    std::string digits("0123456789");
    int base = 10;
    
    std::string buffer;	
    
    if(i == 0){
      buffer += "0";
    }
    
    while (i > 0){ 
      buffer = digits[i%base] + buffer;
      i	= i/base;
    }	
    return buffer;
  }
  
  bool Encoder::test (std::string modelfile,
                      std::string outfile,
                      std::string testfile)
  {
    
    Tagger tagger;   
    if (! tagger.open (modelfile.c_str())) {
      std::cerr << tagger.what() << std::endl;
      return false;
    }
    
    // Output 
    ostream_wrapper os (outfile.c_str());
    if (! *os) {
      std::cerr << "open failed: " << outfile << std::endl;
      return false;
    }

    // Test data
    istream_wrapper is (testfile.c_str());
    if (! *is) {
      std::cerr << "open failed: " << testfile << std::endl;
      return false;
    }
    
    while (*is) tagger.parse (0, 0, *is, *os);
    return true;
  }

  bool Encoder::learn (ConfigFile config)
  {
    std::cerr << COPYRIGHT << std::endl;
    
    // svnvish: Begin reading parameters from config
    string opt;
    config.readInto(opt, "optimizer", std::string("SMD"));
    
    bool binmodel;
    config.readInto(binmodel, "binmodel", false);
    
    bool txtmodel;
    config.readInto(txtmodel, "txtmodel", false);
    
    size_t freq;
    config.readInto(freq, "freq", (size_t)1);
    
    size_t maxitr;
    config.readInto(maxitr, "maxitr", (size_t)100);
    
    double C;
    config.readInto(C, "regularizer", 4.0);
    
    double tol;
    config.readInto(tol, "tol", 0.0001);
    
    string templfile;
    config.readInto(templfile, "templfile", std::string("template"));
    
    string trainfile;
    config.readInto(trainfile, "trainfile", std::string("train.txt"));

    string testfile;
    config.readInto(testfile, "testfile", std::string("test.txt"));
    
    string modelfile;
    config.readInto(modelfile, "modelfile", std::string("model"));

    string outfile;
    config.readInto(outfile, "outfile", std::string("output.txt"));

    bool doitrtest;
    config.readInto(doitrtest, "doitrtest", false);

    bool dobatchtest;
    config.readInto(dobatchtest, "dobatchtest", false);
    
    // svnvish: end reading parameters from config
    
    EncoderFeatureIndex feature_index;
    std::vector<Tagger* > x;

    std::cout.setf(std::ios::fixed, std::ios::floatfield);
    std::cout.precision(5);
    
    // read features
    if (! feature_index.open (templfile.c_str(), trainfile.c_str()))
      WHAT_ERROR(feature_index.what());

    std::ifstream ifs (trainfile.c_str());
    if (! ifs) WHAT_ERROR("cannot open: " << trainfile.c_str());
    
//     std::cerr << "reading training data from : " 
//               << trainfile << std::endl<<std::flush; 
    // size_t line = 0;
    
    while (ifs) {
      Tagger *_x = new Tagger ();
      _x->open (&feature_index);
      if (! _x->read (ifs)) WHAT_ERROR(_x->what());
      
      if (! _x->empty()) x.push_back (_x);
      else delete _x;
      
      // if (++line % 100 == 0) std::cout << line << ".. " << std::flush;
    }
    
    ifs.close ();
        
    // std::cerr << "\nDone! " << std::endl;
    
    // end reading features
    
    // svnvish: if you want to dump the kitchen sink out
    // std::cout<< config;
    
    // svnvish: If you want to check the active features
    // feature_index.dump();
    
    feature_index.shrink(freq);
    
    std::vector <Fad<double> > alpha(feature_index.size());     // parameter
    std::vector <Fad<double> > expected(feature_index.size());  // gradient
    std::fill(alpha.begin(), alpha.end(), 0.0);

    std::cerr << "Number of sentences: " << x.size() << std::endl;
    
    std::cerr << "Total features:" << feature_index.size()
              << " active features:  " << feature_index.getrealsize()
              << " (freq: " << freq << ')' << std::endl;
    std::cerr << "C (sigma^{2}): " << C << std::endl << std::endl;

    feature_index.set_alpha(&alpha[0]);
    
    double old_f = 1e+100;
    int    converge = 0;
    
    if (opt == "LBFGS"){
      
      // Use LBFGS. Sigh!
      LBFGS lbfgs;
      lbfgs.init(feature_index.size(), 5);
      
      std::cerr << "Using L-BFGS" << std::endl << std::endl;
      
      for (size_t itr = 0; itr < maxitr; ++itr) {
        
        double f = 0.0;
        size_t err = 0;
        size_t all = 0;
        size_t zeroone = 0;
        
        std::fill (expected.begin(), expected.end(), 0.0);
        
        feature_index.clear();
        
        for (size_t i = 0; i < x.size(); ++i) {
          f += x[i]->calcGradient (&expected[0]);
          x[i]->eval (err, all, zeroone);
        }
        
        for (size_t k = 0; k < feature_index.size(); ++k) {
          f += (alpha[k].v() * alpha[k].v() / (2.0 * C));
          expected[k] += alpha[k]/C;
        }
        
        double diff = (itr == 0 ? 1.0 : std::abs(old_f - f)/old_f);
        std::cout << "iter="    << itr
                  << " terr="    << 1.0 * err/all
                  << " serr="    << 1.0 * zeroone/x.size()
                  << " obj=" << f
                  << " diff="   << diff << std::endl;
        old_f = f;
        
        if (diff < tol) converge++; else converge = 0;
        if (itr > maxitr || converge == 3)  break; // 3 is ad-hoc
        

        std::vector<double> myalpha(feature_index.size());     // parameter
        std::vector<double> myexpected(feature_index.size());  // gradient
        
        for (size_t j = 0; j < feature_index.size(); ++j) {
          myalpha[j] = alpha[j].v();
          myexpected[j] = expected[j].v();
        }
        
        int ret = lbfgs.optimize(&myalpha[0], &f, &myexpected[0]);
        
        for (size_t j = 0; j < feature_index.size(); ++j) {
          alpha[j].v() = myalpha[j];          
        }
        
        if (ret <= 0) WHAT_ERROR(lbfgs.what());

        if(doitrtest){
          // svnvish: BUGBUG
          // Ugly hack but I can't find a better way to do it :(
          if (!feature_index.save(modelfile.c_str(), txtmodel)){
            WHAT_ERROR(feature_index.what());
          }
          string tmp_outfile = "itr"+itostr(itr) + "_" + outfile; 
          test(modelfile, tmp_outfile, testfile);
        }
        
      }
      
    } else if (opt == "Collins"){ 

      int    batches;
      config.readInto(batches, "batches");
      
      // Use Collins Perceptron
      size_t xs = x.size();
      size_t fs = feature_index.size();
      size_t bs = xs/batches;
      assert(x.size()%bs == 0);
      
      // svnvish: BUGBUG
      // Misleading name. Here expected is actually 
      // phi(x_{i}, y*) - \phi(x_{i}, y_{i})
      std::fill(expected.begin(), expected.end(), 0.0);
      
      std::cerr << "Using Collins Perceptron" << std::endl << std::endl;
      
      for (size_t itr = 0; itr < maxitr; ++itr) {
        double f = 0.0; 
        size_t err = 0;
        size_t all = 0;
        size_t zeroone = 0;
        feature_index.clear();
        
        int batch = 0;
        for (size_t i = 0; i < xs; i++) {
          
          f += x[i]->calcCollins(&expected[0]);
          x[i]->eval(err, all, zeroone);
          
          for (size_t j = 0; j < fs; ++j) {
            alpha[j].v() -= expected[j].v();
          }
          
          if(dobatchtest && (itr == 0) && 
             ((batch < 100) || (batch % 100 == 0))){
            // svnvish: BUGBUG
            // Ugly hack but I can't find a better way to do it :(
            if (!feature_index.save(modelfile.c_str(), txtmodel)){
              WHAT_ERROR(feature_index.what());
            }
            string tmp_outfile = "batch" + itostr(batch) + "_" + outfile; 
            test(modelfile, tmp_outfile, testfile);
          }
          
          // The notion of a batch is different for Collins perceptron
          // This is just a hack to make reporting consistent 
          if((i % bs) == 0)
            batch++;
          
        }
        
        double diff = (itr == 0 ? 1.0 : std::abs((old_f - f)/old_f));
        
        std::cout << "iter="    << itr
                  << " terr="   << err/(double) all
                  << " serr="   << zeroone/(double) xs
                  << " obj="    << f
                  << " diff="   << diff;
        std::cout << std::endl;
        
        if (diff < tol) converge++; else converge = 0;
        if (itr > maxitr || converge == 3)  break; // 3 is ad-hoc
        
        old_f = f;

        if(doitrtest){
          // svnvish: BUGBUG
          // Ugly hack but I can't find a better way to do it :(
          if (!feature_index.save(modelfile.c_str(), txtmodel)){
            WHAT_ERROR(feature_index.what());
          }
          string tmp_outfile = "itr"+itostr(itr) + "_" + outfile; 
          test(modelfile, tmp_outfile, testfile);
        }

      }
       
    }else {
      
      // Use SMD. Yay!
      double mu;
      double lambda;
      double eta_init;
      int    batches;
      config.readInto(mu, "mu");
      config.readInto(lambda, "lambda");
      config.readInto(eta_init, "eta_init");
      config.readInto(batches, "batches");
      
      size_t bs = x.size(); 
      size_t xs = x.size();
      size_t fs = feature_index.size();
      
      bs = x.size()/batches;
      assert(x.size()%bs == 0);
      
      std::vector<double> eta(fs);
      std::fill(eta.begin(), eta.end(), eta_init);
      std::fill(expected.begin(), expected.end(), 0.0);
      
      std::cerr << "Using SMD with"
                << " eta_0 = "<< eta_init << "," 
                << " mu = " << mu << ","
                << " lambda = " << lambda << ","
                << " batch size = " << bs << std::endl << std::endl;

      for (size_t itr = 0; itr < maxitr; ++itr) {
        double f = 0.0; 
        size_t err = 0;
        size_t all = 0;
        size_t zeroone = 0;
        feature_index.clear();
        
        int batch = 0;
        for (size_t i = 0; i < xs; i += bs) {
          for (size_t j = 0; j < bs; ++j) {
            f += x[i+j]->calcGradient(&expected[0]);
            x[i+j]->eval(err, all, zeroone);
          }
          
          for (size_t j = 0; j < fs; ++j) {
            double& ev = expected[j].v();
            double& ed = expected[j].d();
            double& av = alpha[j].v();
            double& ad = alpha[j].d();
            
            // regularization
            double ee = bs/(C*xs);
            ev += av*ee;
            ed += ad*ee;

            // Function value
            f += av*av*ee/2.0;
            
            // nic: BUGBUG
            // don't use exp, use max
            // eta[j] *= std::exp(-mu*ev*ad);
            eta[j] *= std::max(0.5, 1 - mu*ev*ad);
            alpha[j].v() -= eta[j]*ev;
            ad = lambda*ad - eta[j]*(ev + lambda*ed);

            ev = ed = 0.0;
          }
          
          if(dobatchtest && (itr == 0) && 
             ((batch < 100) || (batch % 100 == 0))){
            // svnvish: BUGBUG
            // Ugly hack but I can't find a better way to do it :(
            if (!feature_index.save(modelfile.c_str(), txtmodel)){
              WHAT_ERROR(feature_index.what());
            }
            string tmp_outfile = "batch" + itostr(batch) + "_" + outfile; 
            test(modelfile, tmp_outfile, testfile);
          }
          
          batch++;
        }
        
        double diff = (itr == 0 ? 1.0 : std::abs((old_f - f)/old_f));
        
        std::cout << "iter="    << itr
                  << " terr="   << err/(double) all
                  << " serr="   << zeroone/(double) xs
                  << " obj="    << f
                  << " diff="   << diff;
        std::cout << std::endl;
        
        if (diff < tol) converge++; else converge = 0;
        if (itr > maxitr || converge == 3)  break; // 3 is ad-hoc
        
        old_f = f;

        if(doitrtest){
          // svnvish: BUGBUG
          // Ugly hack but I can't find a better way to do it :(
          if (!feature_index.save(modelfile.c_str(), txtmodel)){
            WHAT_ERROR(feature_index.what());
          }
          string tmp_outfile = "itr"+itostr(itr) + "_" + outfile; 
          test(modelfile, tmp_outfile, testfile);
        }

      }
    }
    
    if (!feature_index.save(modelfile.c_str(), txtmodel))
      WHAT_ERROR(feature_index.what());
    std::cout << "\nDone! ";
    
    return true;
    
  }
  
}


