/*
  CRF++ -- Yet Another CRF toolkit

  $Id: feature.cpp,v 1.8 2005/10/30 03:54:15 taku Exp $;

  Copyright (C) 2005 Taku Kudo <taku@chasen.org>

  This is free software with ABSOLUTELY NO WARRANTY.

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307	USA
*/
#include "feature_index.h"
#include "node.h"
#include "path.h"

namespace CRFPP
{
  static char *BOS[4] = { "_B-1", "_B-2", "_B-3", "_B-4"};
  static char *EOS[4] = { "_B+1", "_B+2", "_B+3", "_B+4"};

  static char *get_index (char *&p,
			  size_t pos,
			  const std::vector <std::vector <char *> >& x)
  {
    if (*p++ !='[') return 0;

    int col = 0;
    int row = 0;

    int neg = 1;
    if (*p++ == '-') neg = -1;
    else --p;

    for (;*p;++p) {
      switch (*p) {
      case '0': case '1': case '2': case '3': case '4':
      case '5': case '6': case '7': case '8': case '9':
	row = 10 * row + (*p - '0'); break;
      case ',':
	++p; goto NEXT1;
      default: return  0;
      }
    }

  NEXT1:

    for (;*p;++p) {
      switch (*p) {
      case '0': case '1': case '2': case '3': case '4':
      case '5': case '6': case '7': case '8': case '9':
	col = 10 * col + (*p - '0'); break;
      case ']': goto NEXT2;
      default: return 0;
      }
    }

  NEXT2:

    row *= neg;

    if (row < -4 || row > 4 || col < 0 || col >= (int)x[pos].size()) return 0;

    int idx = pos + row;
    if (idx < 0) return BOS[-idx-1];
    if (idx >= (int)x.size()) return EOS[idx - x.size()];

    return x[idx][col];
  }

  static bool apply_rule (std::ostringstream &os,
			  char* p,
			  size_t pos,
			  std::vector <std::vector <char *> >& x)
  {
    os.str(""); // clear
    char *r;

    for (; *p; p++) {
      switch (*p) {
      default:
        os << *p; break;
      case '%':
        switch (*++p) {
        case 'x':
          ++p;
          r = get_index (p, pos, x); if (! r) return false;
          os << r;
          break;
        default:
          return false;
        }
        break;
      }
    }
    
    os << '\0';

    return true;
  }

  // selectively prevent some tokens from contributing to the overall
  // feature vector
  static bool is_kosher(const std::string& token){
    
    
    // Token is kosher only if it does not contain a N or a __nil__ flag. 
    return ((token.rfind(":N") == std::string::npos) && 
            (token.rfind(":__nil__") == std::string::npos));
    
  }
  

  void FeatureIndex::rebuildFeatures (std::vector <std::vector <char *> >& x,
                                      std::vector <std::vector <Node *> >& node)
  {
    path_freelist.free();
    node_freelist.free();
    
    for (size_t cur = 0; cur < x.size(); ++cur) {
      int *f = feature_cache.shift ();
      for (size_t i = 0; i < y.size(); ++i) {
        Node *n = node_freelist.alloc ();
        n->clear ();
        n->x = cur;
        n->y = i;
        n->fvector = f;
        node[cur][i] = n;
      }
    }
    
    for (size_t cur = 1; cur < x.size(); ++cur) {
      int *f = feature_cache.shift ();
      for (size_t j = 0; j < y.size(); ++j) {
        for (size_t i = 0; i < y.size(); ++i) {
          Path *p = path_freelist.alloc ();
          p->clear ();
          p->add (node[cur-1][j], node[cur][i]);
          p->fvector = f;
        }
      }
    }

  }
  
  

#define ADD { int id = this->getID(os.str().data()); if (id != -1) feature.push_back (id); }
  
  bool FeatureIndex::buildFeatures (std::vector <std::vector <char *> >& x,
                                    std::vector <std::vector <Node *> >& node)
  {
    std::ostringstream os;
    std::vector <int> feature;
    typedef std::vector<char *>::iterator tmpl_itr;
    
    for (size_t cur = 0; cur < x.size(); ++cur) {
      
      for (tmpl_itr it = unigram_templs.begin(); 
           it != unigram_templs.end(); ++it) {
        if (! apply_rule (os, *it, cur, x))
          WHAT_FALSE("format error: " << *it);
        
        // svnvish: BUGBUG
        // Does this work?
        if( is_kosher(os.str().data())){
         
          ADD;
        }else{
          // std::cout << "skipping ... " << os.str().data() << std::endl;
        }
        
        // ADD;
        // std::cout<< os.str().data() <<std::endl;
      }
      feature_cache.add (feature);
      feature.clear();
    }
    
    for (size_t cur = 1; cur < x.size(); ++cur) {
      for (tmpl_itr it = bigram_templs.begin(); 
           it != bigram_templs.end(); ++it) {
        if (! apply_rule (os, *it, cur, x))
          WHAT_FALSE("format error: " << *it);
        
        // svnvish: BUGBUG
        // Does this work?
        if( is_kosher(os.str().data())){
          ADD;
        }else{
          // std::cout << "skipping ... " << os.str().data() << std::endl;
        }
        // ADD;
        // std::cout<< os.str().data() <<std::endl;
      }
      feature_cache.add (feature);
      feature.clear();
    }

    return true;
  }
#undef ADD
}
