/*
  CRF++ -- Yet Another CRF toolkit

  $Id: feature_index.h,v 1.8 2005/10/30 03:54:15 taku Exp $;

  Copyright (C) 2005 Taku Kudo <taku@chasen.org>

  This is free software with ABSOLUTELY NO WARRANTY.

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307	USA
*/
#ifndef CRFPP_FEATURE_INDEX_H
#define CRFPP_FEATURE_INDEX_H

#include <iostream>
#include <vector>
#include <map>
#include <sstream>
#include "common.h"
#include "feature_cache.h"
#include "path.h"
#include "node.h"
#include "freelist.h"
#include "mmap.h"
#include "darts.h"

namespace CRFPP
{
  class FeatureIndex
  {
  protected:
    size_t maxid;
    Fad<double> *alpha;
    float  *alpha_float;
    size_t xsize_;
    FeatureCache       feature_cache;
    std::vector<char*> unigram_templs;
    std::vector<char*> bigram_templs;
    std::vector<char*>	y;
    FreeList<Path>  path_freelist;
    FreeList<Node>  node_freelist;
    FreeList<char>  char_freelist;

    virtual int getID (const char *) = 0;

    std::ostringstream _what;

  public:

    static const unsigned int version = MODEL_VERSION;

    size_t size	 () { return maxid; }
    size_t xsize () { return xsize_; }
    size_t ysize () { return y.size(); }
    const char* get_y (size_t i) { return y[i]; }
    void   set_alpha (Fad<double> *_alpha) { alpha = _alpha; }

    char *strdup (const char *);
    void calcCost (Node *);
    void calcCost (Path *);

    bool buildFeatures (std::vector <std::vector <char *> >&,
			std::vector <std::vector <Node *> >&);

    void rebuildFeatures (std::vector <std::vector <char *> >&,
			  std::vector <std::vector <Node *> >&);

    std::string what() { return _what.str(); }

    virtual bool open (const char*, const char*) = 0;
    virtual void clear () = 0;

    FeatureIndex (): maxid(0), alpha(0), alpha_float(0),
		     path_freelist (8192*16),
		     node_freelist (8192),
		     char_freelist(8192) {};
    virtual ~FeatureIndex () {};
  };

  class EncoderFeatureIndex: public FeatureIndex {
  private:
    std::map <std::string, std::pair<int, unsigned int> > dic;
    int getID (const char *);
    bool openTemplate (const char *);
    bool openTagSet   (const char *);
  public:
    bool open(const char*, const char*);
    bool save (const char *, bool);
    bool convert (const char *, const char*);
    void clear ();
    void shrink (size_t) ;
    void dump(void);
    size_t getrealsize(void){ return dic.size(); }
};

  class DecoderFeatureIndex: public FeatureIndex {
  private:
    MeCab::Mmap <char> mmap;
    Darts::DoubleArray da;
    int getID (const char *);
  public:
    bool open (const char *, const char *);
    void clear ();
  };
}

#endif
