/*
  CRF++ -- Yet Another CRF toolkit
 
  $Id: tagger.h,v 1.5 2005/10/30 03:54:15 taku Exp $;

  Copyright (C) 2005 Taku Kudo <taku@chasen.org>

  This is free software with ABSOLUTELY NO WARRANTY.
  
  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.
  
  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.
  
  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#ifndef CRFPP_TAGGER_H
#define CRFPP_TAGGER_H

#include <iostream>
#include <vector>
#include <sstream>
#include <queue>
#include "feature_index.h"

namespace CRFPP 
{
  class Tagger 
  {
  private:
    
    enum { LEARN, TEST };

    int mode;
    FeatureIndex *feature_index;

    size_t ysize;
    std::vector <std::vector <char *> > x;      // input
    std::vector <std::vector <Node *> > node;   // unigram
    std::vector <unsigned short int> answer;
    std::vector <unsigned short int> result;
     
    Fad<double> Z; // log sum
    
    std::ostringstream _what;

    void forwardbackward (); 
    void viterbi   ();
    void buildLattice ();

    // nbest
    struct QueueElement {
      Node *node;
      QueueElement *next;
      Fad<double> fx; // f(x) = h(x) + g(x): cost function for A* search
      Fad<double> gx; // g(x)
    };

    class QueueElementComp {
    public: 
      const bool operator () (QueueElement *q1, QueueElement *q2) { return (q1->fx > q2->fx); }
    };

    bool initNbest();
    bool next (Fad<double> &);

    std::priority_queue <QueueElement*, std::vector <QueueElement *>, QueueElementComp> agenda;
    FreeList <QueueElement> nbest_freelist;

  public:

    Tagger (): mode(LEARN), feature_index (0), ysize (0), nbest_freelist(128) {};
    virtual ~Tagger() { close (); };

    void   eval (size_t&, size_t&, size_t&);
    double calcGradient (Fad<double> *);

    double calcCollins(Fad<double> *);

    bool  open (const char*);    // for test
    bool  open (FeatureIndex *); // for learn
    bool  close ();
    void  clear ();
    bool  empty ();

    std::string what() { return _what.str(); }

    bool read  (std::istream &);
    bool parse (size_t, int, std::istream &, std::ostream &);
  };
}

#endif
